﻿/*
 * ShapeSpot
 * Written 2012 by Sven Killig <sven@killig.de>
 * 
 * in:
 * depth
 * bass
 * treble
 * 
 * out:
 * brightness
 * color
 * transparency
 * 
 * <Window AllowsTransparency="True" Background="Transparent" Foreground="{x:Null}" Topmost="True">
 * this.colorBitmap = new WriteableBitmap(this.sensor.DepthStream.FrameWidth, this.sensor.DepthStream.FrameHeight - 80, 96.0, 96.0, PixelFormats.Bgra32, null);
 * 
 * performance:
 * - debug
 * - winamp preset
 * 
 */


namespace Microsoft.Samples.Kinect.DepthBasics
{
    using System;
    using System.Globalization;
    using System.IO;
    using System.Windows;
    using System.Windows.Media;
    using System.Windows.Media.Imaging;
    using Microsoft.Kinect;
    using NAudio.CoreAudioApi;
    using NAudio.Wave;
    using NAudio.Dsp;

    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window {
        /// <summary>
        /// Active Kinect sensor
        /// </summary>
        private KinectSensor sensor;

        /// <summary>
        /// Bitmap that will hold color information
        /// </summary>
        private WriteableBitmap colorBitmap;

        /// <summary>
        /// Intermediate storage for the depth data received from the camera
        /// </summary>
        private DepthImagePixel[] depthPixels, depthPixelsFirst;
        bool firstRun = true;
//        int runs;
        static int marginX = 160;
        static int marginY = marginX * 10 / 16;
        static byte fade = 10;

        /// <summary>
        /// The sentinel depth value indicating unknown depth.
        /// </summary>
        private const int UnknownDepth = 0;

        /// <summary>
        /// Intermediate storage for the depth data converted to color
        /// </summary>
        private byte[] colorPixels;

        /// <summary>
        /// Initializes a new instance of the MainWindow class.
        /// </summary>
        public MainWindow() {
            InitializeComponent();
        }








        private void LoadWasapiDevicesCombo() {
            MMDeviceEnumerator deviceEnum = new MMDeviceEnumerator();
            MMDeviceCollection deviceCol = deviceEnum.EnumerateAudioEndPoints(DataFlow.Capture, DeviceState.Active);

            //            Collection<MMDevice> devices = new Collection<MMDevice>();

            foreach (MMDevice device in deviceCol) {
                //                devices.Add(device);
                Console.WriteLine(device.ID + " " + device.DeviceFriendlyName + " " + device.FriendlyName);
            }

            //            this.comboDevices.DataSource = devices;
            //            this.comboDevices.DisplayMember = "FriendlyName";
        }



        Complex[] channelData;
        float lastLevel;
        void waveIn_DataAvailable(object sender, WaveInEventArgs e) {
            /*if (this.InvokeRequired)
            {
                //Debug.WriteLine("Data Available");
                this.BeginInvoke(new EventHandler<WaveInEventArgs>(waveIn_DataAvailable), sender, e);
            }
            else
            {*/
            //Debug.WriteLine("Flushing Data Available");
            //                writer.Write(e.Buffer, 0, e.BytesRecorded);
            //            }
            //Console.Write(".");
            //Console.WriteLine(e.BytesRecorded);
            if (e.BytesRecorded > 0) {
                channelData = new Complex[e.BytesRecorded / BytesPerFrame];
                for (int index = 0; index < e.BytesRecorded; index += BytesPerFrame) {
                    //                short sample = (short)((e.Buffer[index + 1] << 8) | e.Buffer[index + 0]);
                    //                float sample32 = sample / 32768f;
                    float sample32;
                    switch (BitsPerSample) {
                        case 32:
                            sample32 = BitConverter.ToSingle(e.Buffer, index);
                            break;
                        default:
                            throw new ArgumentException("Unsupported BitsPerSample");
                    }
                    //                ProcessSample(sample32);
                    channelData[index / BytesPerFrame].X = sample32;
                    channelData[index / BytesPerFrame].Y = 0;
                }
                GetFFTResults();
            }
        }


        /// <summary>
        /// Performs an FFT calculation on the channel data upon request.
        /// </summary>
        /// <param name="fftBuffer">A buffer where the FFT data will be stored.</param>
        public void GetFFTResults() {
            //            int bufferSize = -1;
            //            Complex[] channelDataClone = new Complex[bufferSize];
            //            channelData.CopyTo(channelDataClone, 0);
            binaryExponentitation = (int)Math.Log(channelData.Length, 2);
            FastFourierTransform.FFT(true, binaryExponentitation, channelData);
            for (int i = 0; i < channelData.Length / 2; i++) {
                // Calculate actual intensities for the FFT results.
                channelData[i].X = (float)Math.Sqrt(channelData[i].X * channelData[i].X + channelData[i].Y * channelData[i].Y);
            }
        }




        IWaveIn waveIn;
        private int binaryExponentitation;
        int Channels, BitsPerSample, BytesPerFrame;

        /// <summary>
        /// Execute startup tasks
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void WindowLoaded(object sender, RoutedEventArgs e) {
            LoadWasapiDevicesCombo();
            waveIn = new WasapiCapture(33);
            Channels = waveIn.WaveFormat.Channels;
            BitsPerSample = waveIn.WaveFormat.BitsPerSample;
            BytesPerFrame = Channels * (BitsPerSample / 8);
            Console.WriteLine(Channels + " ch " + BitsPerSample + " bit " + waveIn.WaveFormat.SampleRate + " Hz");
            waveIn.DataAvailable += new EventHandler<WaveInEventArgs>(waveIn_DataAvailable);
            waveIn.StartRecording();








            // Look through all sensors and start the first connected one.
            // This requires that a Kinect is connected at the time of app startup.
            // To make your app robust against plug/unplug, 
            // it is recommended to use KinectSensorChooser provided in Microsoft.Kinect.Toolkit
            foreach (var potentialSensor in KinectSensor.KinectSensors) {
                if (potentialSensor.Status == KinectStatus.Connected) {
                    this.sensor = potentialSensor;
                    break;
                }
            }

            if (null != this.sensor) {
                // Turn on the depth stream to receive depth frames
                this.sensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);

                // Allocate space to put the depth pixels we'll receive
                this.depthPixels = new DepthImagePixel[this.sensor.DepthStream.FramePixelDataLength];
                this.depthPixelsFirst = new DepthImagePixel[this.sensor.DepthStream.FramePixelDataLength];

                // Allocate space to put the color pixels we'll create
                this.colorPixels = new byte[this.sensor.DepthStream.FramePixelDataLength * sizeof(int)];
//                Array.Clear(colorPixels, 0, colorPixels.Length);

                // This is the bitmap we'll display on-screen
                this.colorBitmap = new WriteableBitmap(this.sensor.DepthStream.FrameWidth - 2 * marginX, this.sensor.DepthStream.FrameHeight - 80 - 2 * marginY, 96.0, 96.0, PixelFormats.Bgra32, null);
//                this.colorBitmap = new WriteableBitmap(this.sensor.DepthStream.FrameWidth - 2 * marginX, this.sensor.DepthStream.FrameHeight - 80 - 2 * marginY, 96.0, 96.0, PixelFormats.Bgr32, null);

                // Set the image we display to point to the bitmap where we'll put the image data
                this.Image.Source = this.colorBitmap;

                // Add an event handler to be called whenever there is new depth frame data
                this.sensor.DepthFrameReady += this.SensorDepthFrameReady;

                // Start the sensor!
                try {
                    this.sensor.Start();
                } catch (IOException) {
                    this.sensor = null;
                }
            }

/*            if (null == this.sensor) {
                this.statusBarText.Text = Properties.Resources.NoKinectReady;
            }*/
        }

        /// <summary>
        /// Execute shutdown tasks
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void WindowClosing(object sender, System.ComponentModel.CancelEventArgs e) {
            waveIn.StopRecording();

            if (null != this.sensor) {
                this.sensor.Stop();
            }
        }

        float maxDev = 0;

        /// <summary>
        /// Event handler for Kinect sensor's DepthFrameReady event
        /// </summary>
        /// <param name="sender">object sending the event</param>
        /// <param name="e">event arguments</param>
        private void SensorDepthFrameReady(object sender, DepthImageFrameReadyEventArgs e) {
            using (DepthImageFrame depthFrame = e.OpenDepthImageFrame()) {
                if (depthFrame != null) {
                    // Copy the pixel data from the image to a temporary array
                    if (firstRun) {
                        firstRun = false;
                        depthFrame.CopyDepthImagePixelDataTo(this.depthPixelsFirst);
                    } else {
                        depthFrame.CopyDepthImagePixelDataTo(this.depthPixels);
                        /*if (runs < 30) {
                            runs++;
                            for (int y = 40; y < depthFrame.Height - 40; ++y) {
                                for (int x = depthFrame.Width - 1; x >= 0; --x) {
                                    int depthIndex = x + (y * depthFrame.Width);
                                    short depth = depthPixels[depthIndex].Depth;

                                }
                            }
                        }*/

                    }

                    // Get the min and max reliable depth for the current frame
                    int minDepth = depthFrame.MinDepth;
                    int maxDepth = depthFrame.MaxDepth;

                    // Convert the depth to RGB
                    int colorPixelIndex = 0;
                    /*for (int i = 0; i < this.depthPixels.Length; ++i)
                    {*/



                    float level = 0, derivative=0;
                    if (binaryExponentitation > 0) {
                        /*for (int x = 0; x < Math.Pow(2, binaryExponentitation) / 2; ++x) {
                            int y = (int)(channelData[x].X * 4000);
                            if (y < 0) y = 0;
                            if (y > depthFrame.Height - 40 - 2 * marginY) y = depthFrame.Height - 40 - 2 * marginY;
                            int cIndex = x * 4 + (y * depthFrame.Width * 4);
//                            if (cIndex > 0 && cIndex<colorPixels.Length)
//                            {
                            this.colorPixels[cIndex] = 255;
                            this.colorPixels[cIndex + 1] = 255;
                            this.colorPixels[cIndex + 2] = 255;
//                            }
                        }*/
                        level = channelData[1].X;
                        derivative = level - lastLevel;
                        maxDev = Math.Max(maxDev, derivative);
                        lastLevel = level;
//                        Console.WriteLine(level + ", " + derivative + ", " + maxDev);
                    }

                    byte beat = (byte)(derivative > 0.05 ? 1 : 0);
//                    byte beat = 1;

                    // loop over each row and column of the depth
                    for (int y = 40 + marginY; y < depthFrame.Height - 40 - marginY; ++y) {
                        for (int x = depthFrame.Width - 1 - marginX; x >= marginX; --x) {
                            // calculate index into depth array
                            int depthIndex = x + (y * depthFrame.Width);





                            // Get the depth for this pixel
                            short depth = depthPixels[depthIndex].Depth;
                            short depthFirst = depthPixelsFirst[depthIndex].Depth;
//                            short depth = (short)(depthPixelsFirst[depthIndex].Depth - depthPixels[depthIndex].Depth);

                            // To convert to a byte, we're discarding the most-significant
                            // rather than least-significant bits.
                            // We're preserving detail, although the intensity will "wrap."
                            // Values outside the reliable depth range are mapped to 0 (black).

                            // Note: Using conditionals in this loop could degrade performance.
                            // Consider using a lookup table instead when writing production code.
                            // See the KinectDepthViewer class used by the KinectExplorer sample
                            // for a lookup table example.
//                            byte intensity = (byte)(depth >= minDepth && depth <= maxDepth ? depth : 0);
//                            byte intensity = (byte)(depthPixelsFirst[depthIndex].Depth - depth > 100 ? 128.0 + beat * 127.0 : 0);
//                            byte intensity = (byte)(depth >= minDepth && depth <= maxDepth ? depth : 0);
                            int intensity = depthFirst - depth > 200 ? beat * 255 : 0;
/*                            intensity = intensity == 0 ? intensity-- : 0;
                            intensity = intensity > 0 ? intensity : 0;*/

/*                            // Write out blue byte
                            this.colorPixels[colorPixelIndex++] = 0;//intensity;

                            // Write out green byte
                            this.colorPixels[colorPixelIndex++] = 0;//intensity;

                            // Write out red byte                        
                            this.colorPixels[colorPixelIndex++] = 0;//intensity;*/

                            // We're outputting BGR, the last byte in the 32 bits is unused so skip it
                            // If we were outputting BGRA, we would write alpha here.
//                            ++colorPixelIndex;
                            if (depth > UnknownDepth && depthFirst > UnknownDepth) {
//                                this.colorPixels[colorPixelIndex+3] = (byte)(255 - intensity);
                                if (intensity > 0) this.colorPixels[colorPixelIndex + 3] = 0;
                                else this.colorPixels[colorPixelIndex + 3] += (byte)Math.Min(fade, 255-this.colorPixels[colorPixelIndex + 3]);
                            } else {
//                                if (this.colorPixels[colorPixelIndex + 3] < 255) this.colorPixels[colorPixelIndex + 3] += fade;
                                this.colorPixels[colorPixelIndex+3] = 255;
                            }
                            colorPixelIndex += 4;
//                            this.colorPixels[colorPixelIndex++] = intensity;
                        }
                    }




                    // Write the pixel data into our bitmap
                    this.colorBitmap.WritePixels(
                        new Int32Rect(0, 0, this.colorBitmap.PixelWidth, this.colorBitmap.PixelHeight),
                        this.colorPixels,
                        this.colorBitmap.PixelWidth * sizeof(int),
                        0);
                }
            }
        }
    }
}